/*
 * STATsrv: STAT-service daemon 
 *          part of dim_STAT tool <http://dimitrik.free.fr>
 *
 * (c) 1997-2010 by (dim) Dimitri KRAVTCHUK <dimitri.kravtchuk <at> free.fr>
 *
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA                   *
 ***************************************************************************
 */


/*
    ...
    v.3.9:  -- Oct.2008 - fixed kill chld processes!
    v.3.10: -- May.2009 - added "&" to hidden command check!
                        - fixed possible data oferflow in cmd sizeof()!! 
    v.3.11: -- Jun.2009 - keep only alphanumeric symbols within a command line!!
    v.3.12: updated accepted chars: see check_char() function..
    v.4.0:  -- Mar.2010 - extended syntax statname:port
*/

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>           /* struct timeval */

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

extern int errno;
extern void exit();
void err();
void msg();
int Nread();
int Nwrite();

#define VERSION      "v.4.0 (dim)"
#define ACCESS_NAME  "/etc/STATsrv/access"

void sigpipe() {  signal(SIGPIPE, sigpipe); }

void sigchld()
{
 int sst;

 wait( &sst );
 signal( SIGCHLD, sigchld );
}

int fd= -1;                         /* fd of network socket */
int fd_cmd= -1;

void sigterm()
{
 msg( "*** EXIT: got SIGTERM! ***" );

 //kill all chld procs!!!
 kill_chld();

 if( fd_cmd != -1 )  close( fd_cmd );
 if( fd != -1 )  close( fd );
 kill( 0, SIGTERM );

 exit(0);
}

Usage( name ) 
char *name;
{ fprintf( stderr, "%s %s\nUsage: %s [-p [IP:]port (default:5000)] \n", name, VERSION, name ); exit( -2 ); }



main( argc, argv )
int argc;
char *argv[];
{
 int domain, fromlen;
 short port= 5000;               /* TCP port number */
 int i, cnt;
 char *chp;
 char *IP= NULL;

 struct sockaddr_in peer;
 int peerlen= sizeof(peer);

 for( i= 1; i < argc; i++ )
 {
  if( strcmp( argv[i], "-p" ) == 0 )
  {
   if( (chp= (char *)strchr( argv[++i], ':' )) == NULL )
   {
    port= atoi( argv[i] );
    fprintf( stderr, "\nSTAT-service %s ==> IP:[any] Port:[%d] PID:%d\n", VERSION, port, getpid() );
    continue;
   }

   *chp = '\0';
   IP= argv[i];
   port= atoi( ++chp );
   fprintf( stderr, "\nSTAT-service %s ==> IP:[%s] Port:[%d] PID:%d\n", VERSION, IP, port, getpid() );
   continue;
  }

  Usage( argv[0] );
 }

 if( access( ACCESS_NAME, 0 ) != 0 )  { printf( "Not found access file: %s\n", ACCESS_NAME ); exit( -2 ); }
 setpgrp();

 sinme.sin_port=  htons(port);
 sinme.sin_family= AF_INET;

 if( IP == NULL ) sinme.sin_addr.s_addr= INADDR_ANY;
 else  
   if( (sinme.sin_addr.s_addr= inet_addr( IP )) == (in_addr_t)-1 )
//   if( inet_aton( IP, &(sinme.sin_addr) ) == 0 ) 
   {
    printf( "## Bad IP address: %s\n\n", IP );
    Usage( argv[0] );
   }

 while(1)
 {
  if((fd= socket(AF_INET, SOCK_STREAM, 0)) < 0 )    err("socket");
  if( bind( fd, &sinme, sizeof(sinme) ) >= 0 )  break;
  msg( "-- Bind Error, will try again..." );
  sleep( 30 );
 }

 signal(SIGPIPE, sigpipe);
 signal(SIGCHLD, sigchld);
 signal(SIGTERM, sigterm);

 listen(fd,0);   /* allow a queue of 0 */
 msg( "*** READY for Service! ***" );

 fromlen= sizeof(frominet);
 domain= AF_INET;

 while( 1 )
 {
  if((fd_cmd= accept( fd, &frominet, &fromlen )) < 0 )
  {
   if( errno == EINTR )  continue;
   err("accept");
  }

  if( fork() == 0 )
  {
   if( getpeername( fd_cmd, (struct sockaddr_in *) &peer, &peerlen ) < 0 )  err("getpeername");
   STATsrv( fd_cmd,  inet_ntoa(peer.sin_addr), port );
   close( fd_cmd );

   exit(0);
  }

  close( fd_cmd );
 }
}


//Valid chars within a shell command
check_char( char ch )
{
 static char valid[]= ".,:*/-_ ";
 int i;

 if( isalnum( ch ) )  return( 0 );

 for( i= 0; valid[i] != '\0'; i++ )
  if( valid[i] == ch )  return( 0 );

 return( -1 );
}


//Check and Execute a shell command
FILE *check_command( cmdline, addr, port, fd )
char *cmdline;
char *addr;
short port;
int fd;
{
 FILE *fp;
 char cmd[100], msg[200], buff[512], *chp, *tail;
 char key[20], key_cmd[40], key_fullcmd[100];
 int i, n, access, Ok, found, List= 0;

 /*
  * replace any "suspected" characters whithin a command line by spaces
  * to avoid any other hidden order...
  */
 for( chp= cmdline; *chp != '\0'; chp++ )
  if( check_char( *chp ) != 0 ) 
   *chp= ' ';

 for( cmd[0]= '\0', i= n= 0; i < (sizeof( cmd ) - 2) && cmdline[n] != '\0'; n++ )
 {
  if( cmdline[n] == ' ' )
  {
   if( i > 0 )  break;
   continue;
  }

  cmd[i++]= cmdline[n];
 }

 cmd[i]= '\0';
 tail= &cmdline[n];
 // remove a path from name
 while( (chp= (char *)strchr( cmd, '/' )) != NULL )  strcpy( cmd, chp+1 );

 if( strcmp( cmd, "STAT_LIST" ) == 0 )
 {
  sprintf( msg, "STAT *** LIST COMMAND (%s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  List= 1;
 }

 if( (fp= fopen( ACCESS_NAME, "r" )) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (STAT service not configured: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 for( access= found= 0, Ok= 1; fgets( buff, sizeof(buff), fp ) != NULL; )
 {
  if( buff[0] == '#' )  continue;
  if( strlen( buff ) < 5 )  continue;
  sscanf( buff, "%s%s%s", key, key_cmd, key_fullcmd );

  if( strcmp( key, "access" ) == 0 )
  {
   if( access && Ok )  continue;
   access= 1;
   Ok= 0;
   if( strcmp( addr, key_cmd ) == 0 )  Ok= 1;
   continue;
  }

  if( strcmp( key, "command" ) == 0 )
  {
   access= 0;
   if( !Ok )  continue;

   //check the case cmd:port
   if( (chp= (char *)strchr( key_cmd, ':' )) != NULL )
   {
    if( atoi( chp+1 ) != port )  continue;
    *chp= '\0';
   }

   if( List )
   {
    sprintf( msg, "STAT: %s\n", key_cmd );
    Nwrite( fd, msg, strlen( msg ) );
    continue;
   }

   if( strcmp( cmd, key_cmd ) == 0 )  { found= 1; break; }
  }
 }

 fclose( fp );

 if( List )
 {
  sprintf( msg, "STAT *** LIST END (%s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 if( !found )
 {
  sprintf( msg, "STAT *** BAD COMMAND (no access: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( NULL );
 }

 /* avoid buff overflow */
 strcpy( buff, key_fullcmd );
 strcat( buff, " " );
 strncat( buff, tail, sizeof(buff) - strlen(buff) - 1 );
 buff[ sizeof(buff)-1 ]= '\0';

 if( (fp= popen( buff, "r" )) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (cannot start: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  return( fp );
 }

 if( fgets( buff, sizeof(buff), fp ) == NULL )
 {
  sprintf( msg, "STAT *** BAD COMMAND (not found: %s)\n", cmd );
  Nwrite( fd, msg, strlen( msg ) );
  pclose( fp );
  return( NULL );
 }

 sprintf( msg, "STAT *** OK COMMAND (cmd: %s)\n", cmd );
 Nwrite( fd, msg, strlen( msg ) );
 Nwrite( fd, buff, strlen( buff ) );
 return( fp );
}


/*
 *  In many cases child processes continue their execution 
 *  even(!) if command pipe is broken...
 *  Need an explicit kill here to be sure to stop all staff..
 */

#define P_SIZE 256
#define B_SIZE 400

kill_chld()
{
 FILE *pp;
 char buff[ B_SIZE + 8];
 long pid;
 long plist[ P_SIZE ];
 int  n, i;
 int  Ok;

 pid= getpid();
 Ok= 0;

 if( access( "/etc/STATsrv/bin/chldPID.sh", 0 ) == 0 )
 {
   while( !Ok )
   {
    sprintf( buff, "/etc/STATsrv/bin/chldPID.sh %d", pid );
    pp= popen( buff, "r" );

    for( n= 0; n < P_SIZE && fgets( buff, B_SIZE, pp ) != NULL; n++ )  
    {
     sscanf( buff, "%d", &plist[n] );
     if( plist[n] < 2 )  n--;
    }

    if( n < P_SIZE )  Ok= 1;
    for( i= 0; i < n; i++ )  kill( plist[i], SIGKILL );
    pclose( pp );
   }

  return( 0 );
 }


 if( access( "/etc/STATsrv/bin/ptree", 0 ) == 0 )
 {
  while( !Ok )
  {
   sprintf( buff, "/etc/STATsrv/bin/ptree %d", pid );
   pp= popen( buff, "r" );
 
   while( fgets( buff, B_SIZE, pp ) != NULL )
   {
    sscanf( buff, "%d", &plist[0] );
    if( plist[0] == pid )  break;
   }

   for( n= 0; n < P_SIZE && fgets( buff, B_SIZE, pp ) != NULL; n++ )  
   {
    sscanf( buff, "%d", &plist[n] );
    if( plist[n] < 2 )  n--;
   }

   if( n < P_SIZE )  Ok= 1;
   while( n > 0 )  kill( plist[--n], SIGKILL );
   pclose( pp );
  }
 }

 return( 0 );
}


#define CMDLEN 40

STATsrv( fd, addr, port )
int fd;
char *addr;
short port;
{
 long T;
 int cnt;
 char cmd[ CMDLEN ];
 char buff[1024];
 FILE *pp;

 errno= 0;
 if( (cnt= Nread( fd, buff, sizeof(buff) )) <= 0 )  err( "Nread" );
 time( &T );
 fprintf( stderr, "STATsrv-%d:%d> Connect -> [%s] CMD= \"%s\" TIME: %s", getpid(), port, addr, buff, ctime(&T) );
 fflush( stderr );
 strncpy( cmd, buff, CMDLEN );
 cmd[ CMDLEN-1 ]= '\0';

 if( (pp= check_command( buff, addr, port, fd )) != NULL )
 {
  while( fgets( buff, sizeof( buff ), pp ) != NULL )
   if( Nwrite( fd, buff, strlen( buff ) ) <= 0 )
   {
    // kill child procs!!
    kill_chld();
    pclose( pp );
    break;
   }
 }

 close( fd );
 time( &T );
 fprintf( stderr, "STATsrv-%d:%d> Exit -> [%s] CMD= \"%s\" TIME: %s", getpid(), port, addr, cmd, ctime(&T) );
 fflush( stderr );
}


void err( s )
char *s;
{
 perror(s);
 fprintf( stderr," errno= %d\n", errno );
 fflush( stderr );
 exit( -2 );
}

void msg( s )
char *s;
{
 fprintf( stderr,"STATsrv-%d: %s\n", getpid(), s );
 fflush( stderr );
}

Nread( fd, buf, count )
int fd;
char *buf;
int count;
{
 int cnt;
 cnt= read( fd, buf, count );
 if( cnt >= 0 ) buf[cnt]= 0;
 return( cnt );
}

Nwrite( fd, buf, count )
int fd;
char *buf;
int count;
{
 return( write( fd, buf, count ) );
}

